// -*- C++ -*-
/*!
 * @file FluentBit.h
 * @brief File logger stream class
 * @date $Date$
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2017
 *     Noriaki Ando
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id$
 *
 */

#ifndef RTC_LOGGER_FLUENTBIT_H
#define RTC_LOGGER_FLUENTBIT_H

#include <string>
#include <fstream>
#include <iostream>
#include <fluent-bit.h>

#include <coil/stringutil.h>
#include <rtm/LogstreamBase.h>

#ifndef LINE_MAX
#define LINE_MAX  1024
#endif

#ifndef BUFFER_LEN
#define BUFFER_LEN LINE_MAX
#endif

namespace RTC
{
  /*!
   * @if jp
   * @class FluentbitStream
   *
   * Υ饹 Ϥ fluent-bit 뤿 logstream_buf
   * 饹Ǥ롣FluentBit 饹ˤƥ󥹥󥹲졢
   * SystemLogger  logstream ɲä롣
   *
   * @else
   * @class FluentbitStream
   *
   * This class is a logstream_buf class to send log output to
   * fluent-bit. This is instantiate in the FluentBit class and it is
   * added to SystemLogger's logstreams.
   *
   * @endif
   */
  class FluentBitStream
    : public StreambufType
  {
  public:
    FluentBitStream();

    virtual ~FluentBitStream();

    bool init(const coil::Properties& prop);

    bool createOutputStream(const coil::Properties& prop);

    bool createInputStream(const coil::Properties& prop);

  protected:
    std::streamsize pushLogger();

    virtual std::streamsize xsputn(const char_type* str,
                                   std::streamsize insize);

  private:
    char m_buf[BUFFER_LEN];
    size_t m_pos;

    typedef int FlbHandler;
    FlbHandler m_fibInLib;

    std::vector<FlbHandler> m_flbIn;
    std::vector<FlbHandler> m_flbOut;

    // Static variables
    /*!
     * @brief Fluent-bit context
     */
    static flb_ctx_t* s_flbContext;
    /*!
     * @brief Fluent-bit instance count
     */
    static int s_instance;
  };

  /*!
   * @if jp
   * @class FluentBit
   *
   * Υ饹 Ϥ fluent-bit 뤿Υȥ꡼
   * ѥץ饰󥯥饹Ǥ롣
   *
   * fluent-bit ϥʬۥߥɥ륦 fluentd CǤ롣
   * fluent-bit/fluentd ͡ʥץȥǥμե륿󥰡
   * ԤȤǤ롣Υ饹ϡȥ꡼Υץ饰
   *  FluentBit 饹 std::stream_buff 饹Υ֥饹
   * ꡢºݤ FluentBit ؤΥνʬô饹Ǥ롣
   *
   * ǥեȤǤϡOpenRTMΥϤ (input) ȤƼꡢ
   * rtc.conf ꤵ줿 (output) ФƥФ뤳Ȥ
   * 롣input  fluent-bit ѤǤץ饰 rtc.conf ͭ
   * ˤ뤳ȤǤ¾ fluentd/fluent-bit ΥϤ
   * ꡢCPṲʤɤϤȤƼ뤳ȤǽǤ
   * 롣¼Ūˡޥɥ饤ץ fluent-bit ȤۤƱ
   * ¸ǽˤʤäƤ롣
   *
   * ץϡŪˤ fluent-bit  key-value Υץѥƥ
   * rtc.conf ǻꤹ뤳ȤǤ٤ƤΥץ饰ѤǤ뤬ʲˡ
   * ɽŪʥץ饰ȤΥץ򼨤
   *   
   * * Available Output plugins
   * - reference: http://fluentbit.io/documentation/0.8/output/index.html
   *
   * ** forward: fluentd forwarding
   * ______________________________________________________________________
   * |  key   |                  Description                 |   Default  |
   * ----------------------------------------------------------------------
   * | host   | Target host where Fluent-Bit  or Fluentd are |  127.0.0.1 |
   * |        | listening for Forward messages.              |            |
   * ----------------------------------------------------------------------
   * | port   | TCP port of the target service.              |      24224 |
   * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   * Example:
   * logger.logstream.fluentd.output0.plugin: forward
   * logger.logstream.fluentd.output0.tag:    <tagname>
   * logger.logstream.fluentd.output0.host:   <fluentd_hostname>
   * logger.logstream.fluentd.output0.port:   <fluentd_port>
   *
   * ** es: Elasticsearch
   * ______________________________________________________________________
   * |  key   |                  Description                 |   Default  |
   * ----------------------------------------------------------------------
   * | host   | IP address or hostname of the target         |  127.0.0.1 |
   * |        | Elasticsearch instance.                      |            |
   * ----------------------------------------------------------------------
   * | port   | TCP port of the target Elasticsearch         |       9200 |
   * |        | instance.                                    |            |
   * ----------------------------------------------------------------------
   * | index  | Elastic index.                               | fluentbit  |
   * ----------------------------------------------------------------------
   * | type   | Elastic type.                                | test       |
   * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   *
   * Example:
   * logger.logstream.fluentd.output0.plugin: es
   * logger.logstream.fluentd.output0.tag:    <tagname>
   * logger.logstream.fluentd.output0.host:   <es_hostname>
   * logger.logstream.fluentd.output0.port:   <es_port>
   * logger.logstream.fluentd.output0.index:  <es_index>
   * logger.logstream.fluentd.output0.type:   <es_type>
   *
   * ** http: HTTP POST request in MessagePack format
   * ______________________________________________________________________
   * |   key  |            Description                       |   Default  |
   * ----------------------------------------------------------------------
   * |  Host  | IP address or hostname of the target HTTP    |  127.0.0.1 |
   * |        | Server.                                      |            |
   * ----------------------------------------------------------------------
   * |  Port  | TCP port of the target HTTP Server.          |         80 |
   * ----------------------------------------------------------------------
   * |  Proxy | Specify an HTTP Proxy. The expected format   |            |
   * |        | of this value is http://host:port.           |            |
   * |        | Note that https is not supported yet.        |            |
   * ----------------------------------------------------------------------
   * |  URI   | Specify an optional HTTP URI for the target  |          / |
   * |        | web server, e.g: /something                  |            |
   * ----------------------------------------------------------------------
   * | Format | Specify the data format to be used in the    |    msgpack |
   * |        | HTTP request body, by default it uses        |            |
   * |        | msgpack, optionally it can be set to json.   |            |
   * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   *
   * Example:
   * logger.logstream.fluentd.output0.plugin: http
   * logger.logstream.fluentd.output0.tag:    <tagname>
   * logger.logstream.fluentd.output0.host:   127.0.0.1
   * logger.logstream.fluentd.output0.port:   80
   * logger.logstream.fluentd.output0.proxy:
   * logger.logstream.fluentd.output0.uri:     /openrtm/
   * logger.logstream.fluentd.output0.format:  msgpack
   *
   * ** nats: NATS output plugin
   * ______________________________________________________________________
   * |  key   |                  Description                 |   Default  |
   * ----------------------------------------------------------------------
   * | host   | IP address or hostname of the NATS Server.   |  127.0.0.1 |
   * ----------------------------------------------------------------------
   * | port   | TCP port of the target NATS Server.          |       4222 |
   * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   *
   * Example:
   * logger.logstream.fluentd.output0.plugin: nats
   * logger.logstream.fluentd.output0.tag:    <tagname>
   * logger.logstream.fluentd.output0.host:   <nats_host>
   * logger.logstream.fluentd.output0.port:   <nats_port>
   *
   * * stdout: Standard Output plugin
   *
   * Example:
   * logger.logstream.fluentd.output0.plugin: stdin
   *
   * @else
   * @class FluentbitStream
   *
   * @endif
   */
  class FluentBit
    : public LogstreamBase
  {
  public:
    /*!
     * @if jp
     *
     * @brief 󥹥ȥ饯
     *
     * @else
     *
     * @brief Constructor
     *
     * @endif
     */
    FluentBit();

    /*!
     * @if jp
     *
     * @brief ǥȥ饯
     *
     * @else
     *
     * @brief Destructor
     *
     * @endif
     */
    virtual ~FluentBit(void);

    /*!
     * @if jp
     * @brief 
     *
     * Logstream饹γƼԤ饹ǤϡͿ줿
     * PropertiesɬפʾƳƼԤ
     *
     * @param prop 
     *
     * @else
     *
     * @brief Initializing configuration
     *
     * This operation would be called to configure in initialization.
     * In the concrete class, configuration should be performed
     * getting appropriate information from the given Properties data.
     *
     * @param prop Configuration information
     *
     * @endif
     */
    virtual bool init(const coil::Properties& prop);

    /*!
     * @if jp
     * @brief basic_strembuf ؤΥݥ󥿤֤
     *
     * LoggerǻѤ basic_streambuf ؤΥݥ󥿤֤
     *
     * @return basic_streambuf (coil::LogStreambuf) ؤΥݥ
     *
     * @else
     *
     * @brief Returns a pointer to the basic_streambuf 
     *
     * This operation would returns a pointer to the basic_streambuf
     * or its subclass that is kept in this class.
     *
     * @return pointer to the basic_streambuf (coil::LogStreambuf)
     *
     * @endif
     */
    virtual StreambufType* getStreamBuffer();

  protected:
    FluentBitStream m_fbstream;
  };
}; // namespace RTC

extern "C"
{
  void DLL_EXPORT FluentBitInit();
};

#endif // RTC_LOGGER_FLUENTBIT_H
